/* Problema del canotaje unisex.
 * Autor: Erwin Meza Vega
 * Este programa se encuentra protegido por la licencia GPL
 * http://www.fsf.org/licensing/licenses/gpl.html
 *
 * El algoritmo implementado en este programa toma ideas de los siguientes algoritmos de sincronizacion:
 * * Los Filosofos
 * * Los Lectores / Escritores
 * * El Barbero Dormilon
 * */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <semaphore.h>

#define N 15 //Numero de mujeres
#define M 2 //Numero de hombres
#define MAXBOATS 5 //Numero de botes
#define MAXLOAD 6 //Numero de pasajeros por bote

/* Estructuras para facilitar la manipulacion de hilos POSIX*/
typedef void (*thread_start_func)(void * arg);

typedef struct {
  pthread_t id;
  pthread_attr_t attr;
  thread_start_func start_func;
  void * arg;
}Thread;

/* Funcion auxiliar para facilitar la creacion de hilos POSIX */
Thread * start_thread(thread_start_func func, void * arg); 

/* Declaracion de semaforos y mutex*/
sem_t mutex; //Exclusion mutua
sem_t w_mutex; //Mutex para mujeres
sem_t w_boarding; //Semaforo para las mujeres que estan abordando un bote
sem_t w_unboarding; //Semaforo para las mujeres que se estan bajando del bote
sem_t m_mutex; //Mutex para hombres
sem_t m_boarding; //Semaforo para los hombres que estan abordando un bote
sem_t m_unboarding; //Semaforo para los hombres que se estan bajando del bote
sem_t s_boat[MAXBOATS]; //Semaforo para cada bote
int s_state[MAXBOATS]; //Estado de cada bote
int s_load[MAXBOATS]; //Numero de pasajeros que lleva el bote
int s_gender[MAXBOATS]; //Genero de los pasajeros del bote
sem_t river; //Semaforo para maximo de botes en el rio


/* Arreglo de hilos para mujeres y hombres */
Thread * t_women[N];
Thread * t_men [M];

/* Arreglo de hilos para los botes */

Thread * t_boats [MAXBOATS];

/* Arreglo de id para mujeres y hombres. El id se asigna en la rutina main.*/
int id_women[N];
int id_men[N];

/* Arreglo de id para los botes. El id se asigna en la rutina main. */
int id_boat[MAXBOATS];

/* Variables globales para llevar la cuenta de las mujeres y los hombres.*/
int w_count; //Conteo de mujeres
int w_remaining; //Numero faltante de mujeres
int w_load; //Numero permitido de mujeres ( = MAXLOAD excepto para el ultimo bote de mujeres) 

int m_count; //Conteo de hombres
int m_remaining; //Numero faltante de hombres
int m_load; //Numero permitido de hombres ( = MAXLOAD excepto para el ultimo bote de hombres) 

int w_leader; //Ultima mujer de cada grupo
int m_leader; //Ultimo hombre de cada grupo


/* Funcion auxiliar para facilitar la creacion de hilos POSIX.
 * Entrada: funcion a ejecutar en el hilo, argumento a la funcion
 * Retorna: Apuntador a el hilo creado*/
Thread * start_thread(thread_start_func func, void * arg) {
  Thread* t;
  t=(Thread*)malloc(sizeof(Thread));
  t->start_func=func;
  t->arg = arg;
  pthread_attr_init(&(t->attr));
  pthread_create(&(t->id), &(t->attr), (void*)t->start_func, t->arg);
    return t;
}

/* Funcion para registrar que un hilo desea esperar hasta la
 * terminacion de otro hilo */
void wait_for_thread(Thread *t) {
	if (t == NULL) return;
	pthread_join(t->id, NULL);
}

/* Alias de la funcion sem_wait */
void down(sem_t *sem) {
  sem_wait(sem);
}

/* Alias de la funcion sem_post */
void up(sem_t *sem) {
  sem_post(sem);
}


/* Funcion para buscar el siguiente bote disponible 
 * Retorna el id del bote disponible, o -1 si no hay bote disponible*/
int get_available_boat(){
	printf("\tGet available boat..\n");
	int i;
	for (i=0; i<MAXBOATS; i++) {
	         printf("\t\tBoat %d: %s\n", i, (s_state[i]==0)?"Available":"Busy");
		if (s_state[i] == 0) //bote disponible!!
			return i;
	}
	return -1; //Peligro!!
}

/* Codigo para la mujer 
 * Recibe como parametro un identificador (entero) */
void woman(void *param) {
  int i;
  int j;
  int k;
  i = *((int *)(param)); //Id de la mujer: parametro que llega del main
  printf("Woman %d arrived\n", i);
  down(&w_mutex);
    down(&mutex);
      w_count = w_count + 1;
      w_remaining = w_remaining - 1;
      //printf("\t[%d] Remaining: %d\n", i, w_remaining);
      if (w_count == w_load) { //Se lleno el cupo
	   printf("(Woman %d) Women team ready!. Count = %d\n", i, w_count);
           for (j=0; j<w_count; j++) {
	          up(&w_boarding);
           }
      }else if(w_remaining == 0) { //Ultimo grupo!
	      printf("(Woman %d) Last Women team!!. Count = %d\n", i, w_count);
	      w_load = w_count;
              for (j=0; j<w_count; j++) {
	          up(&w_boarding);
	      }
      }else {
  	up(&w_mutex);
      }
   up(&mutex);

   printf("\tWoman %d waiting to board...\n", i);
   down(&w_boarding);

   down(&mutex);
      w_count = w_count - 1;
      //printf("w_count after %d: %d\n", i, w_count);
      if (w_count == 0) {
	      /* La ultima mujer del grupo solicita el bote!*/
	      printf("\tWoman %d requested travel!\n", i);
	      k = get_available_boat();
	      if (k < 0) {
		      printf("No available boats!!!!!\n");
	      }
	      s_state[k] = 1; // 1 = boarding
	      s_load[k] = w_load; // Cuantos pasajeros lleva el bote
	      s_gender[k] = 0; //0 = mujer, 1 = hombre
	      up(&s_boat[k]); //La ultima mujer del grupo mantiene w_mutex
      }
   up(&mutex);
   down(&w_unboarding);
   down(&mutex);
      w_count = w_count + 1;
      if (w_count == w_load) { //La ultima en bajarse del bote habilita las demas mujeres
	      printf("(Woman %d) was the last to unboard group of %d women!\n", i, w_count);
	      up(&w_mutex); //Deja avanzar a las demas mujeres
	      w_count = 0;
      }
   up(&mutex);
   
  //printf("\tWoman %d exiting\n", i);
}

/* Codigo para el hombre
 * Recibe como parametro un identificador (entero) */
void man(void *param) {
  int i;
  int j;
  int k;
  i = *((int *)(param)); //Id del hombre: parametro que llega del main
  printf("Man %d arrived\n", i);
  down(&m_mutex);
    down(&mutex);
      m_count = m_count + 1;
      m_remaining = m_remaining - 1;
      //printf("\t[%d] Remaining: %d\n", i, m_remaining);
      if (m_count == m_load) { //Se lleno el cupo
	   printf("(Man %d) Men team ready!. Count = %d\n", i, m_count);
           for (j=0; j<m_count; j++) {
	          up(&m_boarding);
           }
      }else if(m_remaining == 0) { //Ultimo grupo!!
	      printf("(Man %d) Last Men team!!. Count = %d\n", i, m_count);
	      m_load = m_count;
              for (j=0; j<m_count; j++) {
	          up(&m_boarding);
	      }
      }else {
  	up(&m_mutex);
      }
   up(&mutex);

   printf("\tMan %d waiting to board...\n", i);
   down(&m_boarding);

   down(&mutex);
      m_count = m_count - 1;
      //printf("m_count after %d: %d\n", i, m_count);
      if (m_count == 0) {
	      /* El ultimo hombre del grupo solicita el bote!*/
	      printf("\tMan %d requested travel!\n", i);
	      k = get_available_boat();
	      if (k < 0) {
		      printf("No available boats!!!!!\n");
	      }
	      s_state[k] = 1; // 1 = boarding
	      s_load[k] = m_load; // Cuantos pasajeros lleva el bote
	      s_gender[k] = 1; //0 = mujer, 1 = hombre
	      up(&s_boat[k]); //El ultimo hombre del grupo mantiene w_mutex
      }
   up(&mutex);
   down(&m_unboarding);
   down(&mutex);
      m_count = m_count + 1;
      if (m_count == m_load) { //El ultimo en bajarse del bote habilita los demas hombres
	      printf("(Man %d) was the last to unboard group of %d men!\n", i, m_count);
	      up(&m_mutex); //Deja avanzar a los demas hombres
	      m_count = 0;
      }
   up(&mutex);
   
  //printf("\tMan %d exiting\n", i);
}

/*
 * Codigo para el bote
 * Recibe como parametro el identificador del bote (entero)
 * */
void boat(void *param) {
  int i;
  int j;
  i = *((int*)(param));
  printf("Boat %d created\n", i);
  while (1) { 
    printf("Boat %d ready.\n", i);
    down(&s_boat[i]);
    printf("Boat %d [%s]requesting river...\n", i, (s_gender[i] == 0)?"Women":"Men");
    down(&river);
    printf("Boat %d traveling\n", i);
    sleep(rand()%5); //Simular un tiempo aleatorio de viaje
    printf("Boat %d finished travel!\n", i);
    down(&mutex);
    for (j=0; j<s_load[i]; j++) {
	    if (s_gender[i] == 0) { //Mujeres bajandose?
	       up(&w_unboarding);
	    }else { //No, hombres bajandose!
	       up(&m_unboarding);
	    }
    }
    s_load[i] = 0; //Bote vacio
    s_state[i] = 0; //Bote disponible
    up(&mutex);
    up(&river); //Ya realizo el recorrido, listo para otro grupo!
  }
  printf("End of day for boat %d\n");
}


/* Rutina que se encarga de crear los hilos */

int main(int argc, char * argv[]) {
	
  int i;
  int nwomen; //Para el identificador de las mujeres
  int nmen; // Para el identificador de los hombres
  int nboats; // Para el identificador de los botes
  Thread *t;

  /* Inicializar los contadores para los Id de los hilos*/
  nwomen = 0;
  nmen = 0;
  nboats = 0;

  /* Inicializar las variables globales */
  w_remaining = N; //Inicialmente faltan N mujeres por viajar
  w_count = 0; //y han llegado 0
  
  
  m_remaining = M; //Inicialmente faltan M hombres por viajar
  m_count = 0; //y han llegado 0

  w_load = MAXLOAD;
  m_load = MAXLOAD;
  

  printf("Main Started\n");
  /* Inicializar la semilla aleatoria */
  srand(time(NULL));

  /* Inicializar semaforos y mutex */

  sem_init(&mutex, 0, 1);

  sem_init(&w_mutex, 0, 1);
  sem_init(&w_boarding, 0, 0);
  sem_init(&w_unboarding, 0, 0);

  sem_init(&m_mutex, 0, 1);
  sem_init(&m_boarding, 0, 0);
  sem_init(&m_unboarding, 0, 0);
  sem_init(&river, 0, 2);

 //Inicializar los semaforos de los botes
  for (i=0; i<MAXBOATS; i++) {
    sem_init(&s_boat[i], 0, 0);
  }

  /* Primero se crean los botes, aunque no es estrictamente necesario hacerlo
   * antes de que lleguen las mujeres y los hombres, la sincronizacion
   * deberia garantizar que si las mujeres o los hombres intentan abordar un
   * bote y este no se encuentra disponible , se bloquean hasta que el bote este disponible */
  
  for (i=0; i<MAXBOATS; i++) {
	id_boat[i]=i;
	t = start_thread(boat, (void*)&id_boat[i]);
	t_boats[i] = t;
	s_state[i] = 0; // 0 = ready
  }
  
  /*.. ahora llegan las mujeres y los hombres... Se crean hilos en forma
   * aleatoria hasta que se hayan creado todos los hilos para las mujeres y
   * todos los hilos para los hombres */

  while (nwomen < N ||  nmen < M) {
      if (rand() % 2 == 0 && nwomen < N) { //Crear mujeres
        id_women[nwomen]=nwomen;
        t = start_thread(woman, (void*)&id_women[nwomen]);
	t_women[nwomen] = t;
        nwomen = nwomen + 1;
      }else if (nmen < M) { //Crear hombres
        id_men[nmen]=nmen;
        t = start_thread(man, (void*)&id_men[nmen]);
	t_men[nmen] = t;
        nmen = nmen + 1;
      }
  }

  //printf("Created %d women and  %d men\n", nwomen, nmen);

  /* Ahora hacer que el hilo main espere por la terminacion de todos los hilos */

  /*.. esperar que terminen todas las mujeres... */
  for (i=0; i<N; i++) {
	wait_for_thread(t_women[i]);  
  }

  /*.. esperar que terminen todos los hombres... */
  for (i=0; i<M; i++) {
	wait_for_thread(t_men[i]);  
  }

  /* No se debe esperar a que terminen los botes, porque estos se bloquean en su
   * semaforo y el programa no terminara nunca...*/

  
  /*.. esperar que terminen todos los botes... 
   * Si se habilita esta espera, al final habra que presionar CTRL-D para que el
   * programa termine...
   * */
  /*for (i=0; i<MAXBOATS; i++) {
	wait_for_thread(t_boats[i]);  
  }*/

  printf("Main Finished\n");

  return 0;
}
